/*
 * qmail-destroy: Remove messages from a qmail queue.
 * Copyright (C) 2015 LinuxMagic
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author(s): William Storey <william@linuxmagic.com>
 *
 * $Id: qmail-destroy.c 19985 2015-10-15 22:41:05Z william $
 *
 * This program provides a way to delete a message from the queue.
 */

#include <errno.h>
#include <getopt.h>
#include "queue-remove.h"
#include "queue-remove-util.h"
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// path to queue by default.
#define DEFAULT_QUEUE_ROOT "/var/qmail/queue"

// number of directories queue is split into by default.
#define DEFAULT_DIR_SPLIT 23

// user to try to run as if not root.
#define RUN_AS_USER "qmails"

// group to run as if not root.
// because the directories are typically set up so as
// qmail group needed to read - and qmails user isn't enough.
#define RUN_AS_GROUP "qmail"

// command line arguments.
struct __args {
    // path to the root of the queue. usually /var/qmail/queue
    char * queue_path;

    // the number of directories the queue is split into.
    int dir_split;

    // the message to delete. this is the unique identifier
    // that is also the inode number.
    int message_id;

    // whether to actually delete - otherwise only report.
    bool perform_delete;
};

// forward declarations
static void __usage(char * const *);
static struct __args * __get_args(const int, char * const *);
static void __destroy_args(struct __args *);

//! program entry.
/*!
 * @params int Number of arguments
 * @params char** Argument array
 *
 * @returns int 0 if success, non-zero otherwise.
 *
 * NOTE: being a little less const-y on argv for getopt().
 */
int main(const int argc, char * const * argv)
{
    struct __args * arg = NULL;
    int message_exists = 0;
    int removed_count = 0;

    arg = __get_args(argc, argv);
    if (!arg) {
        // we should have output already.
        return 1;
    }

    // you must run as a particular user & group.
    // otherwise modifying queue files is not possible.
    if (!qru_set_run_as_group(RUN_AS_GROUP)) {
        // we should have output already.
        return 1;
    }
    if (!qru_set_run_as(RUN_AS_USER)) {
        // we should have output already.
        return 1;
    }

    // check if it's in the queue.
    // this is nice to do first so we know that at least at this point
    // it is in the queue. else it is not definitive whether we can say
    // the message was removed or not.
    message_exists = qmail_is_message_in_queue(arg->queue_path,
            arg->dir_split, (long unsigned int) arg->message_id);
    if (message_exists == -1) {
        printf("Unable to check if message is in the queue\n");
        __destroy_args(arg);
        return 1;
    }

    // we should report already if there are permission problems.
    if (message_exists == 0) {
        printf("Message is not in the queue.\n");
        __destroy_args(arg);
        return 1;
    }

    if (!arg->perform_delete) {
        printf("Message found. If you want to delete it then please pass the -d flag.\n");
        // TODO: we could show some information about the message?
        //   subject? from?
        __destroy_args(arg);
        return 0;
    }

    // remove it.
    removed_count = qmail_queue_remove(arg->queue_path, arg->dir_split,
            (long unsigned int) arg->message_id);
    if (removed_count < 0) {
        printf("Unable to remove message from the queue\n");
        __destroy_args(arg);
        return 1;
    }

    printf("Removed message %d (%d recipients flagged)\n", arg->message_id,
        removed_count);
    __destroy_args(arg);

    return 0;
}

//! write out usage.
/*!
 * @params char** Argument array
 *
 * @returns void
 *
 * NOTE: being a little less const-y on argv for getopt().
 */
static void __usage(char * const * argv)
{
    if (!argv || !argv[0] || strlen(argv[0]) == 0) {
        printf("__usage(): %s\n", strerror(EINVAL));
        return;
    }
    printf("Delete a message from the queue.\n"
           "\n"
           "Usage: %s <arguments>\n"
           "\n"
           "Arguments may be:\n"
           "    -h              : Usage information\n"
           "\n"
           "    [-p <path>]     : Qmail queue path. Defaults to /var/qmail/queue.\n"
           "\n"
           "    [-s <number>]   : The number of directories your queue is split into.\n"
           "                      Defaults to 23 (/var/qmail/queue/mess/0 ... /var/qmail/queue/mess/22).\n"
           "\n"
           "    -m <number>     : The unique message id in the queue.\n"
           "                      This is the numeric base filename of the\n"
           "                      message.\n"
           "\n"
           "    [-d]            : Delete the message from the queue.\n"
           "                      If not given we will only report what we\n"
           "                      would otherwise do.\n"
           "\n",
       argv[0]);
}

//! parse and validate arguments.
/*!
 * @params int Number of arguments
 * @params char** Argument array
 *
 * @returns struct __args*
 *
 * NOTE: being a little less const-y on argv for getopt().
 */
static struct __args * __get_args(const int argc, char * const * argv)
{
    struct __args * arg = calloc(1, sizeof(struct __args));
    if (!arg) {
        printf("__get_args(): %s\n", strerror(ENOMEM));
        return NULL;
    }

    // explicitly set our defaults. while the struct is zero'd anyway
    // and some of these will be this way no matter what let's be
    // explicit for clarity.
    arg->queue_path = strdup(DEFAULT_QUEUE_ROOT);
    if (!arg->queue_path) {
        printf("%s\n", strerror(ENOMEM));
        return NULL;
    }
    arg->dir_split = DEFAULT_DIR_SPLIT;
    arg->message_id = -1;
    arg->perform_delete = false;

    // retrieve our arguments.
    int opt = -1;
    while ((opt = getopt(argc, argv, "hp:s:m:d")) != -1) {
        switch (opt) {
            case 'p':
                if (arg->queue_path) {
                    free(arg->queue_path);
                    arg->queue_path = NULL;
                }
                arg->queue_path = strdup(optarg);
                if (!arg->queue_path) {
                    printf("__get_args(): %s\n", strerror(ENOMEM));
                    __destroy_args(arg);
                    return NULL;
                }
                break;
            case 's':
                arg->dir_split = atoi(optarg);
                break;
            case 'm':
                arg->message_id = atoi(optarg);
                break;
            case 'h':
                __usage(argv);
                __destroy_args(arg);
                return NULL;
                break;
            case 'd':
                arg->perform_delete = true;
                break;
            default:
                printf("Unknown option %c\n", opt);
                __usage(argv);
                __destroy_args(arg);
                return NULL;
                break;
        }
    }

    // ensure we have the arguments we need.
    if (!arg->queue_path || strlen(arg->queue_path) == 0) {
        printf("You must provide a queue path.\n");
        __usage(argv);
        __destroy_args(arg);
        return NULL;
    }
    if (!qru_directory_exists(arg->queue_path)) {
        printf("Queue path does not exist: %s\n", arg->queue_path);
        __usage(argv);
        __destroy_args(arg);
        return NULL;
    }

    if (arg->dir_split <= 0) {
        printf("Invalid directory split value. Must be greater than zero.\n");
        __usage(argv);
        __destroy_args(arg);
        return NULL;
    }

    // I believe message id must be > 0
    if (arg->message_id <= 0) {
        printf("Invalid message id. Must be greater than zero.\n");
        __usage(argv);
        __destroy_args(arg);
        return NULL;
    }

    return arg;
}

//! destroy argument struct
/*!
 * @params struct __args*
 *
 * @returns void
 */
static void __destroy_args(struct __args * arg)
{
    if (!arg) {
        return;
    }
    if (arg->queue_path) {
        free(arg->queue_path);
    }
    free(arg);
}
